import sys, string
import sdk32

text = open('sdk32.txt').read()
module = sdk32.parse('module', text)

buildvalue_funcs = {}

def need_buildvalue(name):
    buildvalue_funcs[name] = 1

def python_member_type(name, ptr):
    if name == 'int':
        return 'T_INT'
    if name == 'char':
        if ptr:
            return 'T_STRING'
        return 'T_CHAR'
    if name == 'short':
        return 'T_SHORT'
    if name == 'byte':
        return 'T_BYTE'
    return ''

def python_parse_type(name, ptr):
    if not ptr and name in ('short', 'char', 'byte'):
        return 'int'
    return name

def python_type(name):
    for type in module.type_list:
        if name == type.name:
            return 'equiv', type.equiv, type.ptr
    for struct in module.struct_list:
        if name == struct.name:
            return 'struct', '%s_Obj' % struct.name, 0
    for callback in module.callback_list:
        if name == callback.name:
            return 'callback', 'PyObject', 1
    return 'native', name, 0

def find_struct(name):
    for struct in module.struct_list:
        if struct.name == name:
            return struct

def find_sub_structs():
    for struct in module.struct_list:
        for elem in struct.elem_list:
            tclass, tname, tptr = python_type(elem.type)
            if tclass == 'struct':
                need_buildvalue(elem.type)

def obj_has_attr(obj, name):
    for attr in obj.attr_list:
        if attr.name == name:
            return 1
    return 0

def obj_get_attr(obj, name):
    for attr in obj.attr_list:
        if attr.name == name:
            return attr.value
    raise KeyError('name')

def struct_tag(name):
    for struct in module.struct_list:
        if name == struct.name:
            return struct.tag

def callback_name(name):
    return '%s_Callback' % name

def is_void_type(tname):
    return tname == 'void'

########################################################################
# Function Documentation Section
# ------------------------------
# Builds LaTeX output which looks like:
# 
# \begin{funcdesc}{BeginPaint}{wnd}
# Returns \var{ret} and \var{ps}.
# \end{funcdesc}
def func_latex(func):
    parts = []
    parts.append('\\begin{funcdesc}{')
    parts.append(func.name)
    parts.append('}{')
    num = 0
    for arg in func.arg_list:
        if obj_has_attr(arg, 'out') or obj_has_attr(arg, 'cons'):
            continue
        if num:
            parts.append(', ')
        num = num + 1
        parts.append(arg.name)
    parts.append('}\n')
    parts.append('Returns')
    tclass, tname, tptr = python_type(func.ret)
    rets = []
    if not is_void_type(tname):
        rets.append('ret')
    for arg in func.arg_list:
        if not obj_has_attr(arg, 'out'):
            continue
        rets.append(arg.name)
    for i in range(len(rets)):
        if i > 0:
            parts.append(',')
            if i == len(rets) - 1:
                parts.append(' and')
        parts.append(' \\var{%s}' % rets[i])
    parts.append('.\n')
    parts.append('\\end{funcdesc}\n')
    return string.join(parts, '')

def func_latex_section():
    parts = []
    for func in module.func_list:
        parts.append(func_latex(func))
        parts.append('\n')
    return string.join(parts, '')

########################################################################
# Function Body Section
# ---------------------
# Builds functions which look like:
#
#static char sdk32_BeginPaint__doc__[] = "BeginPaint(wnd) -> ret, ps";
#
#/* HDC BeginPaint(HWND wnd, PAINTSTRUCT *ps) */
#static PyObject *sdk32_BeginPaint(PyObject *module, PyObject *args)
#{
#    int wnd;
#    PAINTSTRUCT tmp_ps;
#    HDC ret;
#
#    if (!PyArg_ParseTuple(args, "i", &wnd))
#        return NULL;
#
#    ret = BeginPaint((HWND)wnd, &tmp_ps);
#
#    return Py_BuildValue("iO&", (int)ret, buildvalue_PAINTSTRUCT, &tmp_ps);
#}
#
# Note that this should be the first section evaluated since it
# establishes some global variables which control the generation of
# the other sections.


def buildvalue_char(name):
    tclass, tname, tptr = python_type(name)
    if tclass == 'struct':
        return 'O&'
    elif tclass in ('equiv', 'native'):
        if tname == 'int':
            return 'i'
        elif tname == 'char':
            if tptr:
                return 's'
            else:
                return 'c'
        elif tname != 'void':
            raise ValueError('unrecognised type %s' % name)
    elif tclass == 'callback':
        return 'O'
    elif not is_void_type(tname):
        raise ValueError('unrecognised type %s' % name)

#static char sdk32_BeginPaint__doc__[] = "";
def func_doc(func):
    parts = []
    parts.append('static char ')
    parts.append(module.id)
    parts.append('_')
    parts.append(func.name)
    parts.append('__doc__[] = "')
    parts.append(func.name)
    parts.append('(')
    num = 0
    for arg in func.arg_list:
        if obj_has_attr(arg, 'out') or obj_has_attr(arg, 'cons'):
            continue
        if num:
            parts.append(', ')
        num = num + 1
        parts.append(arg.name)
    parts.append(')')
    tclass, tname, tptr = python_type(func.ret)
    done_arrow = 0
    if not is_void_type(tname):
        parts.append(' -> ')
        parts.append('ret')
        done_arrow = 1
    for arg in func.arg_list:
        if not obj_has_attr(arg, 'out'):
            continue
        if done_arrow:
            parts.append(', ')
        else:
            parts.append(' -> ')
            done_arrow = 1
        parts.append(arg.name)
    parts.append('";\n')
    return string.join(parts, '')

# BOOL GetClientRect(HWND wnd, RECT *rect)
def func_prototype(func):
    parts = []
    parts.append(func.ret)
    parts.append(' ')
    parts.append(func.name)
    parts.append('(')
    num = 0
    for arg in func.arg_list:
        if num:
            parts.append(', ')
        num = num + 1
        parts.append(arg.type)
        parts.append(' ')
        if arg.ptr:
            parts.append('*')
        parts.append(arg.name)
    parts.append(')')
    return string.join(parts, '')

#/* BOOL GetClientRect(HWND wnd, RECT *rect) */
def func_comment(func):
    parts = []
    parts.append('/* ')
    parts.append(func_prototype(func))
    parts.append(' */\n')
    return string.join(parts, '')

class Decls:
    def __init__(self):
        self.types = {}
        self.verbatim = []

    def add(self, type, name):
        if not self.types.get(type):
            self.types[type] = []
        self.types[type].append(name)

    def add_verbatim(self, decl):
        self.verbatim.append(decl)

    def decl_list(self):
        decls = []
        for type, vars in self.types.items():
            decls.append('%s %s' % (type, string.join(vars, ', ')))
        for decl in self.verbatim:
            decls.append(decl)
        return decls

#    int wnd;
#    RECT_Obj *rect;
#    BOOL ret;
#
def func_vars(func):
    parts = []
    need_sep = 0
    decls = Decls()
    for arg in func.arg_list:
        if obj_has_attr(arg, 'decl'):
            decls.add_verbatim(obj_get_attr(arg, 'decl'))
        if obj_has_attr(arg, 'cons'):
            continue
        tclass, tname, tptr = python_type(arg.type)
        if obj_has_attr(arg, 'argconv') or obj_has_attr(arg, 'null'):
            decls.add('PyObject', '*tmp_%s' % arg.name)
        if obj_has_attr(arg, 'decl'):
            pass
        elif obj_has_attr(arg, 'out') and tclass == 'struct':
            decls.add(arg.type, 'tmp_%s' % arg.name)
        elif tptr or arg.ptr:
            if obj_has_attr(arg, 'null'):
                if tptr:
                    decls.add(arg.type, arg.name)
                else:
                    decls.add(arg.type, '*%s' % arg.name)
            else:
                decls.add(python_parse_type(tname, 1), '*%s' % arg.name)
        else:
            decls.add(python_parse_type(tname, 0), arg.name);
    tclass, tname, tptr = python_type(func.ret)
    if not is_void_type(tname):
        decls.add(func.ret, 'ret')

    for decl in decls.decl_list():
        parts.append('    ')
        parts.append(decl)
        parts.append(';\n')
        need_sep = 1
    if need_sep:
        parts.append('\n')
    return string.join(parts, '')

#    if (!PyArg_ParseTuple(args, "iO!", &wnd, &RECT_Type, &rect))
#        return NULL;
#
def func_parse_args(func):
    parts = []
    parts.append('    ')
    parts.append('if (!PyArg_ParseTuple(args, "')
    for arg in func.arg_list:
        if obj_has_attr(arg, 'out') or obj_has_attr(arg, 'cons'):
            continue
        tclass, tname, tptr = python_type(arg.type)
        if tclass == 'callback' \
             or obj_has_attr(arg, 'argconv') or obj_has_attr(arg, 'null'):
            parts.append('O')
        elif tclass == 'struct':
            parts.append('O!')
        elif tname in ('int', 'short', 'byte'):
            parts.append('i')
        elif tname == 'char':
            if tptr:
                parts.append('s')
            else:
                parts.append('i')
        elif tname == 'void':
            parts.append('O')
        else:
            raise ValueError('unrecognised type %s' % tname)
    parts.append('"')
    for arg in func.arg_list:
        if obj_has_attr(arg, 'out') or obj_has_attr(arg, 'cons'):
            continue
        tclass, tname, tptr = python_type(arg.type)
        if obj_has_attr(arg, 'argconv') or obj_has_attr(arg, 'null'):
            parts.append(', &tmp_%s' % arg.name)
        else:
            if tclass == 'struct':
                parts.append(', &%s_Type' % arg.type)
            parts.append(', &%s' % arg.name)
    parts.append('))\n')
    parts.append('        return NULL;\n')
    for arg in func.arg_list:
        if obj_has_attr(arg, 'out') or obj_has_attr(arg, 'cons'):
            continue
        tclass, tname, tptr = python_type(arg.type)
        if tclass == 'callback':
            parts.append('    if (')
            parts.append(arg.name)
            parts.append(' != Py_None && !PyCallable_Check(')
            parts.append(arg.name)
            parts.append(')) {\n')
            parts.append('        PyErr_SetString(PyExc_ValueError, "')
            parts.append(arg.name)
            parts.append(' must be callable or None");\n')
            parts.append('        return NULL;\n')
            parts.append('    }\n')
        elif obj_has_attr(arg, 'argconv'):
            parts.append('    if (!')
            parts.append(obj_get_attr(arg, 'argconv'))
            parts.append('("')
            parts.append(arg.name)
            parts.append('", tmp_')
            parts.append(arg.name)
            parts.append(', &')
            parts.append(arg.name)
            parts.append('))\n')
            parts.append('        return NULL;\n')
        elif obj_has_attr(arg, 'null'):
            parts.append('    if (tmp_')
            parts.append(arg.name)
            parts.append(' == Py_None)\n')
            parts.append('        ')
            parts.append(arg.name)
            parts.append(' = NULL;\n')
            if tclass == 'struct':
                struct = find_struct(arg.type)
                parts.append('    else if (tmp_')
                parts.append(arg.name)
                parts.append('->ob_type == &')
                parts.append(arg.type)
                parts.append('_Type)\n')
                parts.append('        ')
                parts.append(arg.name)
                parts.append(' = &((')
                parts.append(arg.type)
                parts.append('_Obj *)tmp_')
                parts.append(arg.name)
                parts.append(')->')
                parts.append(struct.tag)
                parts.append(';\n')
            else:
                parts.append('    else if (PyString_Check(tmp_')
                parts.append(arg.name)
                parts.append('))\n')
                parts.append('        ')
                parts.append(arg.name)
                parts.append(' = PyString_AsString(tmp_')
                parts.append(arg.name)
                parts.append(');\n')
            parts.append('    else {\n')
            parts.append('        PyErr_SetString(PyExc_TypeError, "')
            if tclass == 'struct':
                parts.append(arg.type)
            else:
                parts.append('string')
            parts.append(' or None is required");\n')
            parts.append('        return NULL;\n')
            parts.append('    }\n')
    parts.append('\n')
    return string.join(parts, '')

#    code from input file
#
def func_code(func):
    parts = []
    for line in func.code_list:
        parts.append('    ')
        parts.append(line)
        parts.append('\n')
    return string.join(parts, '')

#    ret = GetClientRect((HWND)wnd, &rect->rect);
#
def func_call(func):
    parts = []
    parts.append('    ')
    tclass, tname, tptr = python_type(func.ret)
    if not is_void_type(tname):
        parts.append('ret = ')
    parts.append(func.name)
    parts.append('(')
    num = 0
    for arg in func.arg_list:
        if num:
            parts.append(', ')
        num = num + 1
        if obj_has_attr(arg, 'cons'):
            parts.append(obj_get_attr(arg, 'cons'))
            continue
        if obj_has_attr(arg, 'null'):
            parts.append(arg.name)
            continue
        tclass, tname, tptr = python_type(arg.type)
        if tclass == 'struct':
            if obj_has_attr(arg, 'out'):
                parts.append('&tmp_')
                parts.append(arg.name)
            else:
                parts.append('&')
                parts.append(arg.name)
                parts.append('->')
                parts.append(struct_tag(arg.type))
        elif tclass == 'equiv':
            parts.append('(%s)%s' % (arg.type, arg.name))
        elif tclass == 'callback':
            parts.append(callback_name(arg.type))
        else:
            parts.append(arg.name)
    parts.append(');\n')
    return string.join(parts, '')

#    if (debug_file != NULL)
#        debug_msg("GetClientRect(wnd=0x%08x, rect=0x%08x)\n", wnd, rect);
#
def func_debug_call(func):
    parts = []
    parts.append('    if (debug_file != NULL)\n')
    parts.append('        debug_msg("')
    parts.append(func.name)
    parts.append('(')
    num = 0
    for arg in func.arg_list:
        if num:
            parts.append(', ')
        num = num + 1
        if obj_has_attr(arg, 'cons'):
            parts.append(obj_get_attr(arg, 'cons'))
            continue
        tclass, tname, tptr = python_type(arg.type)
        parts.append(arg.name)
        if tclass == 'struct':
            parts.append('=0x%08x')
        elif tclass == 'equiv':
            parts.append('=0x%08x')
        elif tclass == 'callback':
            parts.append(callback_name(arg.type))
        else:
            parts.append('=0x%08x')
    parts.append(')')
    ret_bc = buildvalue_char(func.ret)
    if ret_bc:
        parts.append(' -> 0x%08x')
    parts.append('\\n"')
    for arg in func.arg_list:
        if obj_has_attr(arg, 'cons'):
            continue
        tclass, tname, tptr = python_type(arg.type)
        if tclass == 'callback':
            continue
        parts.append(', ')
        parts.append(arg.name)
    if ret_bc:
        parts.append(', ret')
    parts.append(');\n')
    parts.append('\n')
    return string.join(parts, '')

#    return PyInt_FromLong((int)ret);
def func_return(func):
    parts = []
    parts.append('    return Py_BuildValue("')
    ret_bc = buildvalue_char(func.ret)
    if ret_bc:
        parts.append(ret_bc)
    for arg in func.arg_list:
        if not obj_has_attr(arg, 'out'):
            continue
        bc = buildvalue_char(arg.type)
        if bc:
            parts.append(bc)
    parts.append('"')
    if ret_bc:
        parts.append(', ')
        tclass, tname, tptr = python_type(func.ret)
        if tclass == 'equiv':
            parts.append('(%s)' % tname)
        parts.append('ret')
    for arg in func.arg_list:
        if not obj_has_attr(arg, 'out'):
            continue
        parts.append(', ')
        tclass, tname, tptr = python_type(arg.type)
        if tclass == 'equiv':
            if tptr:
                parts.append('(%s *)' % tname)
            else:
                parts.append('(%s)' % tname)
        elif tclass == 'struct':
            need_buildvalue(arg.type)
            parts.append('buildvalue_')
            parts.append(arg.type)
            parts.append(', &tmp_')
        parts.append(arg.name)
    parts.append(');\n')
    return string.join(parts, '')

#static PyObject *sdk32_GetClientRect(PyObject *module, PyObject *args)
#{
#    ...
#}
def func_body(func):
    parts = []
    parts.append(func_doc(func))
    parts.append('\n')
    parts.append(func_comment(func))
    parts.append('static PyObject *')
    parts.append(module.id)
    parts.append('_')
    parts.append(func.name)
    parts.append('(PyObject *module, PyObject *args)\n')
    parts.append('{\n')
    parts.append(func_vars(func))
    parts.append(func_parse_args(func))
    if func.code_list:
        parts.append(func_code(func))
    else:
        parts.append(func_call(func))
    parts.append('    if (PyErr_Occurred())\n')
    parts.append('        return NULL;\n\n')
    parts.append(func_debug_call(func))
    parts.append(func_return(func))
    parts.append('}\n')
    return string.join(parts, '')

def func_body_section():
    parts = []
    for func in module.func_list:
        parts.append(func_body(func))
        parts.append('\n')
    return string.join(parts, '')

########################################################################
# Py_BuildValue converter
# -----------------------
# Builds Py_BuildValue converter functions for extension types which
# look like:
#
#PyObject *buildvalue_RECT(RECT* rect)
#{
#    RECT_Obj *self;
#
#    self = PyObject_NEW(RECT_Obj, &RECT_Type);
#    if (self == NULL)
#        return NULL;
#
#    memcpy(&self->rect, rect, sizeof(self->ps));
#
#    return (PyObject *)self;
#}
def buildvalue_func(name):
    parts = []
    parts.append('PyObject *buildvalue_')
    parts.append(name)
    parts.append('(')
    parts.append(name)
    parts.append('* ')
    tag = struct_tag(name)
    parts.append(tag)
    parts.append(')\n{\n    ')
    parts.append(name)
    parts.append('_Obj *self;\n\n    self = PyObject_NEW(')
    parts.append(name)
    parts.append('_Obj, &')
    parts.append(name)
    parts.append('_Type);\n    if (self == NULL)\n        return NULL;\n\n')
    parts.append('    memcpy(&self->')
    parts.append(tag)
    parts.append(', ')
    parts.append(tag)
    parts.append(', sizeof(self->')
    parts.append(tag)
    parts.append('));\n')
    struct = find_struct(name)
    for elem in struct.elem_list:
        tclass, tname, tptr = python_type(elem.type)
        if tclass == 'callback' or (tname == 'char' and tptr):
            parts.append('    real_')
            parts.append(elem.rename)
            parts.append(' = NULL;\n')
        if obj_has_attr(elem, 'init'):
            parts.append('    ');
            parts.append(obj_get_attr(elem, 'init'))
            parts.append(';\n')
    parts.append('    return (PyObject *)self;\n}\n')
    return string.join(parts, '')

########################################################################
# Extension Types Section
# -----------------------
# Defines extension types for all of the structures which look like:
#
#typedef struct {
#    PyObject_HEAD
#    MSG msg;
#} MSG_Obj;
#
#static void MSG_dealloc(MSG_Obj *self)
#{
#    PyMem_DEL(self);
#}
#
#static struct memberlist MSG_memberlist[] = {
#    { "wnd", T_INT, offsetof(MSG, hwnd) },
#    { "message", T_INT, offsetof(MSG, message) },
#    { "wparam", T_INT, offsetof(MSG, wParam) },
#    { "lparam", T_INT, offsetof(MSG, lParam) },
#    { "time", T_INT, offsetof(MSG, time) },
#    { "pt", POINT_Obj, offsetof(MSG, pt) }, /* for dir() only */
#    { NULL }
#};
#
#static PyObject *MSG_getattr(MSG_Obj *self, char *name)
#{
#    if (strcmp(name, "pt") == 0)
#        return buildvalue_POINT(&self->msg.pt);
#    return PyMember_Get((char *)&self->msg, MSG_memberlist, name);
#}
#
#static int MSG_setattr(MSG_Obj *self, char *name, PyObject *v)
#{
#    if (strcmp(name, "pt") == 0) {
#        if (v->ob_type != &POINT_Type) {
#            PyErr_SetString(PyExc_TypeError, "POINT is required");
#            return -1;
#        }
#        memcpy(&self->msg.pt, &((POINT_Obj *)v)->point, sizeof(self->msg.pt));
#        return 0;
#    }
#    return PyMember_Set((char *)&self->msg, MSG_memberlist, name, v);
#}
#
#static char MSG__doc__[] = "Wrapper for MSG structure";
#
#PyTypeObject MSG_Type = {
#    PyObject_HEAD_INIT(0)
#    0,                          /*ob_size*/
#    "MSG",                      /*tp_name*/
#    sizeof(MSG_Obj),            /*tp_basicsize*/
#    0,                          /*tp_itemsize*/
#    /* methods */
#    (destructor)MSG_dealloc,    /*tp_dealloc*/
#    (printfunc)0,               /*tp_print*/
#    (getattrfunc)MSG_getattr,   /*tp_getattr*/
#    (setattrfunc)MSG_setattr,   /*tp_setattr*/
#    (cmpfunc)0,                 /*tp_compare*/
#    (reprfunc)0,                /*tp_repr*/
#    0,                          /*tp_as_number*/
#    0,                          /*tp_as_sequence*/
#    0,                          /*tp_as_mapping*/
#    (hashfunc)0,                /*tp_hash*/
#    (ternaryfunc)0,             /*tp_call*/
#    (reprfunc)0,                /*tp_str*/
#
#    /* Space for future expansion */
#    0L, 0L, 0L, 0L,
#    MSG__doc__                 /* Documentation string */
#};

#typedef struct {
#    PyObject_HEAD
#    MSG rect;
#} MSG_Obj;
def struct_decl(struct):
    parts = []
    parts.append('typedef struct {\n    PyObject_HEAD\n    ')
    parts.append(struct.name)
    parts.append(' ')
    parts.append(struct.tag)
    parts.append(';\n')
    for elem in struct.elem_list:
        tclass, tname, tptr = python_type(elem.type)
        if tclass == 'callback' or (tname == 'char' and tptr):
            parts.append('    PyObject *real_');
            parts.append(elem.rename);
            parts.append('; /* python proxy */\n')
    parts.append('} ')
    parts.append(struct.name)
    parts.append('_Obj;\n')
    return string.join(parts, '')

#static void MSG_dealloc(MSG_Obj *self)
#{
#    PyMem_DEL(self);
#}
def struct_dealloc(struct):
    parts = []
    parts.append('static void ')
    parts.append(struct.name)
    parts.append('_dealloc(')
    parts.append(struct.name)
    parts.append('_Obj *self)\n{\n')
    for elem in struct.elem_list:
        tclass, tname, tptr = python_type(elem.type)
        if tclass == 'callback' or (tname == 'char' and tptr):
            parts.append('    Py_XDECREF(self->real_');
            parts.append(elem.rename);
            parts.append(');\n')
    parts.append('    PyMem_DEL(self);\n}\n')
    return string.join(parts, '')

#static struct memberlist MSG_memberlist[] = {
#    { "wnd", T_INT, offsetof(MSG, hwnd) },
#    { "message", T_INT, offsetof(MSG, message) },
#    { "wparam", T_INT, offsetof(MSG, wParam) },
#    { "lparam", T_INT, offsetof(MSG, lParam) },
#    { "time", T_INT, offsetof(MSG, time) },
#    { "pt", POINT_Obj, offsetof(MSG, pt) }, /* for dir() only */
#    { NULL }
#};
def struct_members(struct):
    parts = []
    parts.append('static struct memberlist ')
    parts.append(struct.name)
    parts.append('_memberlist[] = {\n')
    for elem in struct.elem_list:
        tclass, tname, tptr = python_type(elem.type)
        mtype = python_member_type(tname, tptr)
        parts.append('    ')
        parts.append('{ "')
        parts.append(elem.rename)
        parts.append('", ')
        if not mtype:
            parts.append('T_OBJECT')
        else:
            parts.append(mtype)
        parts.append(', offsetof(')
        parts.append(struct.name)
        parts.append(', ')
        parts.append(elem.name)
        parts.append(')')
        if obj_has_attr(elem, 'ro'):
            parts.append(', RO')
        parts.append(' },')
        if not mtype:
            parts.append(' /* for dir() only */')
        parts.append('\n')
    parts.append('    { NULL }\n};\n')
    return string.join(parts, '')

#static PyObject *MSG_getattr(MSG_Obj *self, char *name)
#{
#    if (strcmp(name, "pt") == 0)
#        return buildvalue_POINT(&self->msg.pt);
#    return PyMember_Get((char *)&self->msg, MSG_memberlist, name);
#}
def struct_getattr(struct):
    parts = []
    parts.append('static PyObject *')
    parts.append(struct.name)
    parts.append('_getattr(')
    parts.append(struct.name)
    parts.append('_Obj *self, char *name)\n{\n')
    for elem in struct.elem_list:
        tclass, tname, tptr = python_type(elem.type)
        if tclass == 'struct':
            parts.append('    if (strcmp(name, "')
            parts.append(elem.rename)
            parts.append('") == 0)\n')
            parts.append('        return buildvalue_')
            parts.append(elem.type)
            parts.append('(&self->')
            parts.append(struct.tag)
            parts.append('.')
            parts.append(elem.name)
            parts.append(');\n')
        elif tclass == 'callback' or (tname == 'char' and tptr):
            parts.append('    if (strcmp(name, "')
            parts.append(elem.rename)
            parts.append('") == 0) {\n')
            parts.append('        if (self->real_')
            parts.append(elem.rename)
            parts.append(' == NULL)\n')
            parts.append('            return Py_BuildValue("");\n')
            parts.append('        Py_INCREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('        return self->real_')
            parts.append(elem.rename)
            parts.append(';\n')
            parts.append('    }\n')
    parts.append('    return PyMember_Get((char *)&self->')
    parts.append(struct.tag)
    parts.append(', ')
    parts.append(struct.name)
    parts.append('_memberlist, name);\n}\n')
    return string.join(parts, '')

#static int MSG_setattr(MSG_Obj *self, char *name, PyObject *v)
#{
#    if (strcmp(name, "pt") == 0) {
#        if (v->ob_type != &POINT_Type) {
#            PyErr_SetString(PyExc_TypeError, "POINT is required");
#            return -1;
#        }
#        memcpy(&self->msg.pt, &((POINT_Obj *)v)->point, sizeof(self->msg.pt));
#        return 0;
#    }
#    return PyMember_Set((char *)&self->msg, MSG_memberlist, name, v);
#}
def struct_setattr(struct):
    parts = []
    parts.append('static int ')
    parts.append(struct.name)
    parts.append('_setattr(')
    parts.append(struct.name)
    parts.append('_Obj *self, char *name, PyObject *v)\n{\n')
    for elem in struct.elem_list:
        tclass, tname, tptr = python_type(elem.type)
        if tclass == 'struct':
            other_struct = find_struct(elem.type)
            parts.append('    if (strcmp(name, "')
            parts.append(elem.rename)
            parts.append('") == 0) {\n')
            parts.append('        if (v->ob_type != &')
            parts.append(elem.type)
            parts.append('_Type) {\n')
            parts.append('            PyErr_SetString(PyExc_TypeError, "')
            parts.append(elem.type)
            parts.append(' is required");\n')
            parts.append('            return -1;\n')
            parts.append('        }\n')
            parts.append('        memcpy(&self->')
            parts.append(struct.tag)
            parts.append('.')
            parts.append(elem.name)
            parts.append(', &((')
            parts.append(elem.type)
            parts.append('_Obj *)v)->')
            parts.append(other_struct.tag)
            parts.append(', sizeof(self->')
            parts.append(struct.tag)
            parts.append('.')
            parts.append(elem.name)
            parts.append('));\n')
            parts.append('        return 0;\n')
            parts.append('    }\n')
        elif tclass == 'callback':
            parts.append('    if (strcmp(name, "')
            parts.append(elem.rename)
            parts.append('") == 0) {\n')
            parts.append('        if (v == Py_None) {\n')
            parts.append('            Py_XDECREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('            self->real_')
            parts.append(elem.rename)
            parts.append(' = NULL;\n')
            parts.append('        } else if (!PyCallable_Check(v)) {\n')
            parts.append('            PyErr_SetString(PyExc_ValueError, "')
            parts.append(elem.rename)
            parts.append(' not callable");\n')
            parts.append('            return -1;\n')
            parts.append('        } else {\n')
            parts.append('            Py_XDECREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('            self->real_')
            parts.append(elem.rename)
            parts.append(' = v;\n')
            parts.append('            Py_INCREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('        }\n')
            parts.append('        return 0;\n')
            parts.append('    }\n')
        elif tname == 'char' and tptr:
            parts.append('    if (strcmp(name, "')
            parts.append(elem.rename)
            parts.append('") == 0) {\n')
            parts.append('        if (v == Py_None) {\n')
            parts.append('            Py_XDECREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('            self->real_')
            parts.append(elem.rename)
            parts.append(' = NULL;\n')
            parts.append('        } else if (PyString_Check(v)) {\n')
            parts.append('            Py_XDECREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('            self->real_')
            parts.append(elem.rename)
            parts.append(' = v;\n')
            parts.append('            Py_INCREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('            self->')
            parts.append(struct.tag)
            parts.append('.')
            parts.append(elem.name)
            parts.append(' = PyString_AsString(v);\n')
            parts.append('        } else if (PyInt_Check(v)) {\n')
            parts.append('            Py_XDECREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('            self->real_')
            parts.append(elem.rename)
            parts.append(' = v;\n')
            parts.append('            Py_INCREF(self->real_')
            parts.append(elem.rename)
            parts.append(');\n')
            parts.append('            self->')
            parts.append(struct.tag)
            parts.append('.')
            parts.append(elem.name)
            parts.append(' = (char*)PyInt_AsLong(v);\n')
            parts.append('        } else {\n')
            parts.append('            PyErr_SetString(PyExc_ValueError, "')
            parts.append(elem.rename)
            parts.append(' not string or int");\n')
            parts.append('            return -1;\n')
            parts.append('        }\n')
            parts.append('        return 0;\n')
            parts.append('    }\n')
    parts.append('    return PyMember_Set((char *)&self->')
    parts.append(struct.tag)
    parts.append(', ')
    parts.append(struct.name)
    parts.append('_memberlist, name, v);\n}\n')
    return string.join(parts, '')

#static char MSG__doc__[] = "Wrapper for MSG structure";
#
#PyTypeObject MSG_Type = {
#    PyObject_HEAD_INIT(0)
#    0,                          /*ob_size*/
#    "MSG",                      /*tp_name*/
#    sizeof(MSG_Obj),            /*tp_basicsize*/
#    0,                          /*tp_itemsize*/
#    /* methods */
#    (destructor)MSG_dealloc,    /*tp_dealloc*/
#    (printfunc)0,               /*tp_print*/
#    (getattrfunc)MSG_getattr,   /*tp_getattr*/
#    (setattrfunc)MSG_setattr,   /*tp_setattr*/
#    (cmpfunc)0,                 /*tp_compare*/
#    (reprfunc)0,                /*tp_repr*/
#    0,                          /*tp_as_number*/
#    0,                          /*tp_as_sequence*/
#    0,                          /*tp_as_mapping*/
#    (hashfunc)0,                /*tp_hash*/
#    (ternaryfunc)0,             /*tp_call*/
#    (reprfunc)0,                /*tp_str*/
#
#    /* Space for future expansion */
#    0L, 0L, 0L, 0L,
#    MSG__doc__                 /* Documentation string */
#};
def comment(parts, text):
    col = len(parts[-1])
    if col < 32:
        parts.append(' ' * (32 - col))
    parts.append(text)

def struct_decl_object(struct):
    parts = []
    parts.append('static char ')
    parts.append(struct.name)
    parts.append('__doc__[] = "Wrapper for ')
    parts.append(struct.name)
    parts.append(' structure";\n\n')

    parts.append('PyTypeObject ')
    parts.append(struct.name)
    parts.append('_Type = {\n')
    parts.append('    PyObject_HEAD_INIT(0)\n')
    parts.append('    0,')
    comment(parts, '/*ob_size*/\n')
    parts.append('    "%s",' % struct.name)
    comment(parts, '/*tp_name*/\n')
    parts.append('    sizeof(%s_Obj),' % struct.name)
    comment(parts, '/*tp_basicsize*/\n')
    parts.append('    0,')
    comment(parts, '/*tp_itemsize*/\n')
    parts.append('    /* methods */\n')
    parts.append('    (destructor)%s_dealloc,' % struct.name)
    comment(parts, '/*tp_dealloc*/\n')
    parts.append('    (printfunc)0,')
    comment(parts, '/*tp_print*/\n')
    parts.append('    (getattrfunc)%s_getattr,' % struct.name)
    comment(parts, '/*tp_getattr*/\n')
    parts.append('    (setattrfunc)%s_setattr,' % struct.name)
    comment(parts, '/*tp_setattr*/\n')
    parts.append('    (cmpfunc)0,')
    comment(parts, '/*tp_compare*/\n')
    parts.append('    (reprfunc)0,')
    comment(parts, '/*tp_repr*/\n')
    parts.append('    0,')
    comment(parts, '/*tp_as_number*/\n')
    parts.append('    0,')
    comment(parts, '/*tp_as_sequence*/\n')
    parts.append('    0,')
    comment(parts, '/*tp_as_mapping*/\n')
    parts.append('    (hashfunc)0,')
    comment(parts, '/*tp_hash*/\n')
    parts.append('    (ternaryfunc)0,')
    comment(parts, '/*tp_call*/\n')
    parts.append('    (reprfunc)0,')
    comment(parts, '/*tp_str*/\n')
    parts.append('\n')
    parts.append('    /* Space for future expansion */\n')
    parts.append('    0L, 0L, 0L, 0L,\n')
    parts.append('    %s__doc__' % struct.name)
    comment(parts, '/* Documentation string */\n')
    parts.append('};\n')
    return string.join(parts, '')

#static char sdk32_RECT__doc__[] = "RECT() -> rect";
#
#PyObject *sdk32_RECT()
#{
#    RECT_Obj *self;
#
#    self = PyObject_NEW(RECT_Obj, &RECT_Type);
#    if (self == NULL)
#	return NULL;
#
#    memset(&self->rect, 0, sizeof(self->rect));
#
#    return (PyObject *)self;
#}
def struct_new_func(struct):
    parts = []
    parts.append('static char ')
    parts.append(module.id)
    parts.append('_')
    parts.append(struct.name)
    parts.append('__doc__[] = "')
    parts.append(struct.name)
    parts.append('() -> ')
    parts.append(struct.tag)
    parts.append('";\n\n')
    parts.append('PyObject *')
    parts.append(module.id)
    parts.append('_')
    parts.append(struct.name)
    parts.append('()\n{\n    ')
    parts.append(struct.name)
    parts.append('_Obj *self;\n\n    self = PyObject_NEW(')
    parts.append(struct.name)
    parts.append('_Obj, &')
    parts.append(struct.name)
    parts.append('_Type);\n    if (self == NULL)\n        return NULL;\n\n')
    parts.append('    memset(&self->')
    parts.append(struct.tag)
    parts.append(', 0, sizeof(self->')
    parts.append(struct.tag)
    parts.append('));\n')
    for elem in struct.elem_list:
        tclass, tname, tptr = python_type(elem.type)
        if tclass == 'callback' or (tname == 'char' and tptr):
            parts.append('    self->real_')
            parts.append(elem.rename)
            parts.append(' = NULL;\n')
        if obj_has_attr(elem, 'init'):
            parts.append('    ');
            parts.append(obj_get_attr(elem, 'init'))
            parts.append(';\n')
    parts.append('\n    return (PyObject *)self;\n}\n')
    return string.join(parts, '')

def struct_body(struct):
    parts = []
    parts.append(struct_decl(struct))
    parts.append('\n')
    parts.append(struct_dealloc(struct))
    parts.append('\n')
    parts.append(struct_members(struct))
    parts.append('\n')
    parts.append(struct_getattr(struct))
    parts.append('\n')
    parts.append(struct_setattr(struct))
    parts.append('\n')
    parts.append(struct_decl_object(struct))
    parts.append('\n')
    parts.append(struct_new_func(struct))
    if buildvalue_funcs.has_key(struct.name):
        parts.append('\n')
        parts.append(buildvalue_func(struct.name))
    return string.join(parts, '')

def structs_section():
    parts = []
    for name in module.struct_list:
        parts.append(struct_body(name))
        parts.append('\n')
    return string.join(parts, '')

########################################################################
# Module Header Section
# ---------------------
# Defines start of C source
#
def module_header_section():
    parts = []
    parts.append('#include "Python.h"\n')
    parts.append('#include "structmember.h"\n')
    parts.append('\n')
    for line in module.header_code:
        parts.append(line)
        parts.append('\n')
    return string.join(parts, '')

########################################################################
# Function Table Section
# ----------------------
# Defines all module functions like this:
#
#/* List of methods defined in the module */
#static struct PyMethodDef sdk32_methods[] = {
#    { "EndDialog", (PyCFunction)sdk32_EndDialog, METH_VARARGS, sdk32_EndDialog__doc__ },
#    { NULL } /* sentinel */
#};
def func_table_entry(name):
    parts = []
    parts.append('    { "')
    parts.append(name)
    parts.append('", (PyCFunction)')
    parts.append(module.id)
    parts.append('_')
    parts.append(name)
    parts.append(', METH_VARARGS, ')
    parts.append(module.id)
    parts.append('_')
    parts.append(name)
    parts.append('__doc__ },\n')
    return string.join(parts, '')

def func_table_section():
    parts = []
    parts.append('/* List of methods defined in the module */\n')
    parts.append('static struct PyMethodDef ')
    parts.append(module.id)
    parts.append('_methods[] = {\n')
    parts.append(func_table_entry('set_debug'))
    parts.append(func_table_entry('MAKE_DIALOG_TEMPLATE'))
    parts.append(func_table_entry('gethostbyname'))
    for struct in module.struct_list:
        parts.append(func_table_entry(struct.name))
    for func in module.func_list:
        parts.append(func_table_entry(func.name))
    parts.append('    { NULL } /* sentinel */\n};\n')
    return string.join(parts, '')

########################################################################
# Module Constants Section
# ------------------------
# Defines constants for debugging
#
#enum { CONST_Bool, };
#
#typedef struct {
#    int type;
#    char *name;
#    int value;
#} ValueDesc;
#
#static ValueDesc sdk32_consts[] = {
##ifdef TRUE
#    { CONST_Bool, "TRUE", TRUE },
##endif
#};

def module_const_section():
    parts = []
    parts.append('enum { ')
    num = 0
    for const in module.const_list:
        if num:
            parts.append(', ')
        num = num + 1
        parts.append('CONST_')
        parts.append(const.group)
    parts.append(' };\n\n')
    parts.append('''typedef struct {
    int type;
    char *name;
    int value;
} ValueDesc;

''')
    parts.append('static ValueDesc ')
    parts.append(module.id)
    parts.append('_consts[] = {\n')
    for const in module.const_list:
        for value in const.value_list:
            parts.append('#ifdef ')
            parts.append(value)
            parts.append('\n')
            # { CONST_Bool, "TRUE", TRUE },
            parts.append('    { CONST_')
            parts.append(const.group)
            parts.append(', "')
            parts.append(value)
            parts.append('", ')
            if const.name == 'intstr':
                parts.append('(int)')
            parts.append(value)
            parts.append(' },\n')
            parts.append('#endif\n')
    parts.append('    { 0, NULL, 0 }\n')
    parts.append('};\n\n')
    parts.append('''static int dict_add_int(PyObject *dict, char *key, int value)
{
    int err;
    PyObject *obj = PyInt_FromLong(value);
    if (obj == NULL)
	return 0;

    err = PyDict_SetItemString(dict, key, obj);
    Py_DECREF(obj);
    return !err;
}
''')
    return string.join(parts, '')

########################################################################
# Module Trailer Section
# ----------------------
# Defines end of C source and module initialiser
#
#static char sdk32_module_documentation[] = "";
#
#void initsdk32(void)
#{
#    PyObject *m;
#
#    /* Initialise the type of the new type objects here; doing it here
#     * is required for portability to Windows without requiring C++. */
#    RECT_Type.ob_type = &PyType_Type;
#
#    /* Create the module and add the functions */
#    m = Py_InitModule4("sdk32", sdk32_methods,
#                       sdk32_module_documentation,
#                       (PyObject *)NULL, PYTHON_API_VERSION);
#
#    /* Check for errors */
#    if (PyErr_Occurred()) {
#        char msg[128];
#
#        sprintf(msg, "%s: import failed", module);
#        Py_FatalError(msg);
#    }
#}

def module_trailer_section():
    parts = []
    parts.append('static char ')
    parts.append(module.id)
    parts.append('''_module_documentation[] = "";

void init''')
    parts.append(module.id)
    parts.append('''(void)
{
    PyObject *m, *d;
    ValueDesc *desc;

    /* Initialise the type of the new type objects here; doing it here
     * is required for portability to Windows without requiring C++.
     */
''')
    for struct in module.struct_list:
        parts.append('    ')
        parts.append(struct.name)
        parts.append('_Type.ob_type = &PyType_Type;\n')
    parts.append('''
    /* Create the module and add the functions */
    m = Py_InitModule4("''')
    parts.append(module.id)
    parts.append('", ')
    parts.append(module.id)
    parts.append('''_methods,
                       ''')
    parts.append(module.id)
    parts.append('''_module_documentation,
                       (PyObject *)NULL, PYTHON_API_VERSION);

    d = PyModule_GetDict(m);
    /* Add constants */
    for (desc = ''')
    parts.append(module.id)
    parts.append('''_consts; desc->name != NULL; desc++)
	if (!dict_add_int(d, desc->name, desc->value))
	    break;

    /* Check for errors */
    if (PyErr_Occurred())
        Py_FatalError("''')
    parts.append(module.id)
    parts.append(': import failed");\n}\n')
    return string.join(parts, '')

if __name__ == '__main__':
    find_sub_structs()
    body_text = func_body_section()
    print module_header_section()
    print structs_section()
    print body_text
    print func_table_section()
    print module_const_section()
    print module_trailer_section()
    doc = open('doc/autogen.tex', 'w')
    doc.write(func_latex_section())
    doc.close()
