#include "Python.h"
#include "structmember.h"
#include "windows.h"

PyObject *debug_file = NULL;

char sdk32_set_debug__doc__[] = "set_debug(file) -> old_file";

PyObject *sdk32_set_debug(PyObject *module, PyObject *args)
{
    PyObject *file, *old_file, *tmp;

    if (!PyArg_ParseTuple(args, "O", &file))
        return NULL;

    if (file == Py_None)
	file = NULL;
    else {
	tmp = PyObject_CallMethod(file, "write", "s", "");
	if (tmp == NULL)
	    return NULL;
	Py_DECREF(tmp);
    }

    old_file = debug_file;
    debug_file = file;
    Py_XINCREF(debug_file);

    if (old_file == NULL)
	return Py_BuildValue("");
    return old_file;
}

void debug_msg(char *fmt, ...)
{
    va_list ap;
    char buff[512];
    PyObject *obj;

    if (debug_file == NULL)
	return;

    va_start(ap, fmt);
    vsprintf(buff, fmt, ap);
    va_end(ap);

    obj = PyObject_CallMethod(debug_file, "write", "s", buff);
    Py_XDECREF(obj);
    PyErr_Clear();
    obj = PyObject_CallMethod(debug_file, "flush", "");
    Py_XDECREF(obj);
    PyErr_Clear();
}

typedef struct HWNDFunc {
    struct HWNDFunc *next;
    HWND wnd;
    PyObject *func;
} HWNDFunc;

static HWNDFunc *hwnd_funcs;

static HWNDFunc *hwnd_func_add(HWND wnd, PyObject *func)
{
    HWNDFunc *wf;

    wf = malloc(sizeof(*wf));
    if (wf == NULL)
	return (HWNDFunc*)PyErr_NoMemory();
    wf->wnd = wnd;
    Py_INCREF(func);
    wf->func = func;
    wf->next = hwnd_funcs;
    hwnd_funcs = wf;

    return wf;
}

static HWNDFunc *hwnd_func_find(HWND wnd)
{
    HWNDFunc *wf;

    for (wf = hwnd_funcs; wf; wf = wf->next)
	if (wf->wnd == wnd)
	    return wf;

    return NULL;
}

static void hwnd_func_remove(HWND wnd)
{
    HWNDFunc *wf, *prev;

    for (prev = NULL, wf = hwnd_funcs; wf; prev = wf, wf = wf->next)
	if (wf->wnd == wnd) {
	    if (prev != NULL)
		prev->next = wf->next;
	    else
		hwnd_funcs = wf->next;
	    Py_XDECREF(wf->func);
	    free(wf);
	    return;
	}
}

/* Bounce dialog box messages up into the correct Python function
 */
BOOL CALLBACK dlg_proc_trampoline(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    HWNDFunc *wf;
    BOOL result = 0;

    if (msg == WM_INITDIALOG) {
	wf = hwnd_func_add(wnd, (PyObject *)lparam);
	lparam = 0;
    } else
	wf = hwnd_func_find(wnd);
    if (wf != NULL) {
	PyObject *result_obj = NULL;

	result_obj = PyObject_CallFunction(wf->func, "iiii", wnd, msg,
					   wparam, lparam);
	if (result_obj != NULL && PyInt_Check(result_obj))
	    result = PyInt_AsLong(result_obj);

	Py_XDECREF(result_obj);
	if (result_obj == NULL)
	    EndDialog(wnd, -1);

	if (msg == WM_DESTROY)
	    hwnd_func_remove(wnd);
    }

    return result;
}

typedef struct WNDCLASSFunc {
    struct WNDCLASSFunc *next;
    char *name;
    PyObject *func;
} WNDCLASSFunc;

static WNDCLASSFunc *wndclass_funcs;

int wc_proc_add(const char *name, PyObject *func)
{
    WNDCLASSFunc *wf;

    wf = malloc(sizeof(*wf));
    if (wf == NULL) {
	PyErr_NoMemory();
	return -1;
    }
    wf->name = strdup(name);
    if (wf->name == NULL) {
	free(wf);
	PyErr_NoMemory();
	return -1;
    }
    Py_INCREF(func);
    wf->func = func;
    wf->next = wndclass_funcs;
    wndclass_funcs = wf;

    return 0;
}

static WNDCLASSFunc *wndclass_func_find(char *name)
{
    WNDCLASSFunc *wf;

    for (wf = wndclass_funcs; wf; wf = wf->next)
	if (strcmp(wf->name, name) == 0)
	    return wf;

    return NULL;
}

/* Bounce window messages up into the correct Python function
 */
LRESULT CALLBACK wc_proc_trampoline(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    WNDCLASSFunc *wf;
    int num;
    char name[256];
    LRESULT result = 0;

    num = GetClassName(wnd, name, sizeof(name));

    wf = wndclass_func_find(name);
    if (wf != NULL) {
	PyObject *result_obj = NULL;

	result_obj = PyObject_CallFunction(wf->func, "iiii", wnd, msg,
					   wparam, lparam);
	if (result_obj != NULL && PyInt_Check(result_obj))
	    result = PyInt_AsLong(result_obj);

	Py_XDECREF(result_obj);
    }

    return result;
}

typedef struct TIMERPROCFunc {
    struct TIMERPROCFunc *next;
    HWND wnd;
    UINT timer_id;
    PyObject *func;
} TIMERPROCFunc;

static TIMERPROCFunc *timerproc_funcs;

int timerproc_add(HWND wnd, UINT timer_id, PyObject *func)
{
    TIMERPROCFunc *wf;

    if (func == Py_None)
	return 0;

    wf = malloc(sizeof(*wf));
    if (wf == NULL) {
	PyErr_NoMemory();
	return -1;
    }
    wf->wnd = wnd;
    wf->timer_id = timer_id;
    Py_INCREF(func);
    wf->func = func;
    wf->next = timerproc_funcs;
    timerproc_funcs = wf;

    return 0;
}

void timerproc_remove(HWND wnd, UINT timer_id)
{
    TIMERPROCFunc *wf, *prev;

    for (prev = NULL, wf = timerproc_funcs; wf; prev = wf, wf = wf->next)
	if (wf->wnd == wnd && wf->timer_id == timer_id) {
	    if (prev != NULL)
		prev->next = wf->next;
	    else
		timerproc_funcs = wf->next;
	    Py_XDECREF(wf->func);
	    free(wf);
	    return;
	}
}

static TIMERPROCFunc *timerproc_func_find(HWND wnd, UINT timer_id)
{
    TIMERPROCFunc *wf;

    for (wf = timerproc_funcs; wf; wf = wf->next)
	if (wf->wnd == wnd && wf->timer_id == timer_id)
	    return wf;

    return NULL;
}

void CALLBACK timerproc_trampoline(HWND wnd, UINT msg, UINT event, DWORD time)
{
    TIMERPROCFunc *wf;

    wf = timerproc_func_find(wnd, event);
    if (wf != NULL) {
	PyObject *result_obj;

	result_obj = PyObject_CallFunction(wf->func, "iiii", wnd, msg,
					   event, time);
	Py_XDECREF(result_obj);
    }
}

int conv_nullstr(char *name, PyObject *obj, char **arg)
{
    char msg[128];

    if (obj == Py_None)
	*arg = NULL;
    else if (PyString_Check(obj))
	*arg = PyString_AsString(obj);
    else {
	sprintf(msg, "%s must be string or None", name);
        PyErr_SetString(PyExc_ValueError, msg);
	return 0;
    }
    return 1;
}

int conv_intstr(char *name, PyObject *obj, char **arg)
{
    char msg[128];

    if (PyInt_Check(obj))
	*arg = (char*)PyInt_AsLong(obj);
    else if (PyString_Check(obj))
	*arg = PyString_AsString(obj);
    else {
	sprintf(msg, "%s must be string or int", name);
        PyErr_SetString(PyExc_ValueError, msg);
	return 0;
    }
    return 1;
}

int conv_intnullstr(char *name, PyObject *obj, char **arg)
{
    char msg[128];

    if (obj == Py_None)
	*arg = NULL;
    else if (PyInt_Check(obj))
	*arg = (char*)PyInt_AsLong(obj);
    else if (PyString_Check(obj))
	*arg = PyString_AsString(obj);
    else {
	sprintf(msg, "%s must be string, int or None", name);
        PyErr_SetString(PyExc_ValueError, msg);
	return 0;
    }
    return 1;
}

typedef struct {
    PyObject_HEAD

    PyObject *name;		/* string */
    PyObject *aliases;		/* tuple of strings */
    int addrtype;
    int length;
    PyObject *addr_list;	/* tuple of addresses */
} HOSTENT_Obj;

static void HOSTENT_dealloc(HOSTENT_Obj *self)
{
    Py_XDECREF(self->name);
    Py_XDECREF(self->aliases);
    Py_XDECREF(self->addr_list);
    PyMem_DEL(self);
}

static struct memberlist HOSTENT_memberlist[] = {
    { "name",      T_OBJECT, offsetof(HOSTENT_Obj, name),      RO },
    { "aliases",   T_OBJECT, offsetof(HOSTENT_Obj, aliases),   RO },
    { "addrtype",  T_INT,    offsetof(HOSTENT_Obj, addrtype),  RO },
    { "length",    T_INT,    offsetof(HOSTENT_Obj, length),    RO },
    { "addr_list", T_OBJECT, offsetof(HOSTENT_Obj, addr_list), RO },
    { NULL }
};

static PyObject *HOSTENT_getattr(HOSTENT_Obj *self, char *name)
{
    return PyMember_Get((char *)self, HOSTENT_memberlist, name);
}

static int HOSTENT_setattr(HOSTENT_Obj *self, char *name, PyObject *v)
{
    return PyMember_Set((char *)self, HOSTENT_memberlist, name, v);
}

static char HOSTENT__doc__[] = "Wrapper for HOSTENT structure";

PyTypeObject HOSTENT_Type = {
    PyObject_HEAD_INIT(0)
    0,                          /*ob_size*/
    "HOSTENT",                  /*tp_name*/
    sizeof(HOSTENT_Obj),        /*tp_basicsize*/
    0,                          /*tp_itemsize*/
    /* methods */
    (destructor)HOSTENT_dealloc,/*tp_dealloc*/
    (printfunc)0,               /*tp_print*/
    (getattrfunc)HOSTENT_getattr, /*tp_getattr*/
    (setattrfunc)HOSTENT_setattr, /*tp_setattr*/
    (cmpfunc)0,                 /*tp_compare*/
    (reprfunc)0,                /*tp_repr*/
    0,                          /*tp_as_number*/
    0,                          /*tp_as_sequence*/
    0,                          /*tp_as_mapping*/
    (hashfunc)0,                /*tp_hash*/
    (ternaryfunc)0,             /*tp_call*/
    (reprfunc)0,                /*tp_str*/

    /* Space for future expansion */
    0L, 0L, 0L, 0L,
    HOSTENT__doc__                /* Documentation string */
};

char sdk32_gethostbyname__doc__[] = "gethostbyname(name) -> HOSTENT";

PyObject *sdk32_gethostbyname(PyObject *module, PyObject *args)
{
    char *name;
    HOSTENT *ent;
    HOSTENT_Obj *self;
    PyObject *obj;
    int num;

    if (!PyArg_ParseTuple(args, "s", &name))
        return NULL;

    ent = gethostbyname(name);
    self = PyObject_NEW(HOSTENT_Obj, &HOSTENT_Type);
    if (self == NULL)
        return NULL;

    obj = NULL;
    self->name = self->aliases = self->addr_list = NULL;

    self->name = PyString_FromString(ent->h_name);
    if (self->name == NULL)
	goto error;

    for (num = 0; ent->h_aliases[num] != NULL; num++)
	;
    self->aliases = PyTuple_New(num);
    if (self->aliases == NULL)
	goto error;
    for (num = 0; ent->h_aliases[num] != NULL; num++) {
	obj = PyString_FromString(ent->h_aliases[num]);
	if (obj == NULL)
	    goto error;
	if (PyTuple_SetItem(self->aliases, num, obj))
	    goto error;
	obj = NULL;		/* stolen */
    }

    self->addrtype = ent->h_addrtype;
    self->length = ent->h_length;

    for (num = 0; ent->h_addr_list[num] != NULL; num++)
	;
    self->addr_list = PyTuple_New(num);
    if (self->addr_list == NULL)
	goto error;
    for (num = 0; ent->h_addr_list[num] != NULL; num++) {
	obj = PyInt_FromLong(*(int*)ent->h_addr_list[num]);
	if (obj == NULL)
	    goto error;
	if (PyTuple_SetItem(self->addr_list, num, obj))
	    goto error;
	obj = NULL;		/* stolen */
    }

    return (PyObject *)self;

error:
    Py_XDECREF(self);
    Py_XDECREF(obj);
    return NULL;
}
